"""PyWENO quadrature points.

Requires SymPy.

"""

# Copyright (c) 2011, Matthew Emmett.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.



import numpy as np
import sympy
import sympy.mpmath as mpmath


################################################################################
# polynomial generator, roots etc

def legendre_poly(n):
  '''Return Legendre polynomial P_n(x).

  :param n: polynomial degree
  '''

  x = sympy.var('x')
  p = (1.0*x**2 - 1.0)**n

  top = p.diff(x, n)
  bot = 2**n * 1.0*sympy.factorial(n)
  
  return (top / bot).as_poly()


def find_roots(p):
  '''Return set of roots of polynomial *p*.

  :param p: sympy polynomial

  This uses the *nroots* method of the SymPy polynomial class to give
  rough roots, and subsequently refines these roots to arbitrary
  precision using mpmath.

  Returns a sorted *set* of roots.
  '''

  x = sympy.var('x')
  roots = set()

  for x0 in p.nroots():
    xi = mpmath.findroot(lambda z: p.eval(x, z), x0)
    roots.add(xi)

  return sorted(roots)


################################################################################
# quadrature points


def gauss_legendre(n):
  '''Return Gauss-Legendre nodes.

  Gauss-Legendre nodes are roots of P_n(x).
  '''

  p = legendre_poly(n)
  r = find_roots(p)
  return r


def gauss_lobatto(n):
  """Return Gauss-Lobatto nodes.

  Gauss-Lobatto nodes are roots of P'_{n-1}(x) and -1, 1.
  """

  x = sympy.var('x')
  p = legendre_poly(n-1).diff(x)
  r = find_roots(p)
  r = [mpmath.mpf('-1.0'), mpmath.mpf('1.0')] + r
  return sorted(r)
  

def gauss_radau(n):
  '''Return Gauss-Radau nodes.

  Gauss-Radau nodes are roots of P_n(x) + P_{n-1}(x).
  '''

  p = legendre_poly(n) + legendre_poly(n-1)
  r = find_roots(p)
  return r



"""PyWENO code generation tool kit (kernels)."""

import sympy
import codeprinters

import symbolic

class KernelGenerator(object):
  """Generate kernels for WENO reconstructions.

  Each method generates a 'kernel' for a specific WENO operation:
  computing smoothness indicators, non-linear weights, and
  reconstructions.  These code snippets can, in-turn, be used to
  create seperate functions or larger kernels.

  The pyweno.symbols.names object is used to define naming conventions:

  * smoothness indicators: *sigma*, default ``sigmaX``
  * weights: *omega*, default ``omegaX``
  * intermediate reconstructions: *fr*, default ``frX``.

  For each of the above, the occurance of ``X`` is replaced by the
  left-shift *r*.  For example, for ``k=3`` and ``omega='omegaX'``,
  the weights are stored in ``omega0``, ``omega1``, and ``omega2``,
  each of which are assumed to be in scope.  In some routines the
  accumulator variable ``acc`` is also assumed to be in scope.

  :param lang: ``'c'``, ``'opencl'``, or ``'fortran'``

  """

  def __init__(self, lang, order=None, xi=None, **kwargs):
    import symbols
    if order is not None:
      self.k = (order + 1) / 2
      self.xi = xi
    self.lang = lang.lower()
    self.weights_normalised = False
    symbols.names.lang = lang.lower()


  #############################################################################

  def smoothness(self, reuse=False, beta=None):
    r"""Fully un-rolled smoothness indicator kernel for uniform
    grids.

    The smoothness indicator kernel computes the smoothness indicators
    *sigma* determined by the coefficients in *beta*.  That is:

    .. math::

      \sigma_r = \sum_{m=1}^{2k-1}
                     \sum_{n=1}^{2k-1}
                         \beta_{r,m,n}\, \overline{f}_{i-k+m}\,
                                         \overline{f}_{i-k+n}.

    """

    from symbols import sigma, fmn, f

    kernel = Kernel()

    if beta is None:
      beta = getattr(self, 'beta', symbolic.jiang_shu_smoothness_coefficients(self.k))

    k  = beta.get('k', 0)
    nc = beta.get('l', k)

    if not reuse:
      for r in range(k):
        v = sum([ beta[r,m,n] * f[-r+m] * f[-r+n]
                  for m in range(nc)
                  for n in range(m, nc) ])
        kernel.assign(sigma[r], v)

    else:

      # first pass: compute all unique combinations of f[i-r+m] * f[i-r+n]
      #             usually this would be done for i=k once
      burnin = Kernel()
      burnt  = {}
      for r in range(k):
        for m in range(nc):
          for n in range(m, nc):
            if (-r+m,-r+n) not in burnt:
              burnin.assign(fmn[-r+m,-r+n], f[-r+m] * f[-r+n])
            burnt[-r+m,-r+n] = fmn[-r+m,-r+n]

      # second pass: assuming all unique f[i-r+m] * f[i-r+n] have already been computed,
      #              set/compute all unique f[i+1-r+m] * f[i+1-r+n]
      cache   = {}
      delayed = Kernel()
      for r in range(k):
        for m in range(nc):
          for n in range(m, nc):
            if (-r+m,-r+n) in cache:
              continue
            if (1-r+m,1-r+n) in burnt:
              # set fmn[-r+m,-r+n] from previous pass
              kernel.assign(fmn[-r+m,-r+n], fmn[1-r+m,1-r+n])
            else:
              # compute new fmn[-r+m,-r+n], delay until after all copies are done
              delayed.assign(fmn[-r+m,-r+n], f[-r+m] * f[-r+n])
            cache[-r+m,-r+n] = True

      kernel.src.extend(delayed.src)

      self.burnin = burnin.body()

      # finally, using above compute sigma
      for r in range(k):
        v = sum([ beta[r,m,n] * fmn[-r+m,-r+n]
                  for m in range(nc)
                  for n in range(m, nc) ])
        kernel.assign(sigma[r], v)


    self.beta = beta
    return kernel.body()


  #############################################################################

  def weights(self, varpi=None, split=None, normalise=False, power=2, epsilon='1.0e-6'):
    r"""Fully un-rolled weights kernel for uniform grids.

    The weights kernel computes the weights :math:`\omega^r`
    determined by the smoothness coefficients :math:`\sigma^r` (which
    have already been computed).  The weights :math:`\omega^r` are
    computed from the optimal weights :math:`\varpi^r` according to:

    .. math::

      \omega^r = \frac{\varpi^r}{(\sigma^r + \epsilon)^p}

    The weights are subsequently renormalised (if requested) according
    to:

    .. math::

      \omega^r = \frac{\omega^r}{\sum_j \omega^j}

    :param normalise: re-normalise the weights?
    :param power: power :math:`p` of the denominator
    :param epsilon: :math:`\epsilon`

    If *normalise* is ``False`` the weights are not re-normalised.
    Instead, the re-normalisation occurs during the reconstruction
    step.  This saves a few divisions if the weights are computed
    during the reconstruction.

    """

    from symbols import real, omega, sigma

    kernel = Kernel()

    varpi = getattr(self, 'varpi', None)
    split = getattr(self, 'split', None)
    if varpi is None:
      varpi, split = symbolic.optimal_weights(self.k, self.xi)

    n  = varpi.get('n')
    k  = varpi.get('k')
    nc = varpi.get('l', k)

    scale = { (l,s): sum([ varpi[l,r][s] for r in range(0, k) ])
                                         for l in range(n)
                                         for s in (0, 1) if split[l] }

    self.weights_normalised = normalise

    epsilon = real(epsilon)
    accsym  = real('acc')

    for l in range(n):

      if not split[l]:
        for r in range(0, k):
          kernel.assign(omega[l,r], varpi[l,r] / (sigma[r] + epsilon)**power)
        if normalise:
          kernel.assign(accsym, sum([ omega[l,r] for r in range(0, k) ]))
          for r in range(0, k):
            kernel.assign(omega[l,r], omega[l,r] / accsym)

      else:
        for s, pm in enumerate(('p', 'm')):
          for r in range(0, k):
            kernel.assign(omega[l,r,s], varpi[l,r][s] / scale[l,s] / (sigma[r] + epsilon)**power)
          if normalise:
            kernel.assign(accsym, sum([ omega[l,r,s] for r in range(0, k) ]))
            for r in range(0, k):
              kernel.assign(omega[l,r,s], omega[l,r,s] / accsym)

    self.varpi = varpi
    self.split = split
    return kernel.body()


  #############################################################################

  def reconstruction(self, coeffs=None):
    r"""Fully un-rolled reconstruction kernel for uniform grids.

    The reconstruction kernel computes the WENO reconstruction
    based on the weights *omega* (which have already been
    computed) and the reconstruction coefficients *coeffs*.

    """

    from symbols import omega, fs, fr, f

    kernel = Kernel()

    if coeffs is None:
      coeffs = getattr(self, 'coeffs', symbolic.reconstruction_coefficients(self.k, self.xi))

    n  = coeffs.get('n')
    k  = coeffs.get('k')
    nc = coeffs.get('l', k)

    varpi = getattr(self, 'varpi', None)
    split = getattr(self, 'split', None)
    if varpi is None:
      varpi, split = symbolic.optimal_weights(self.k, self.xi)

    scale = { (l,s): sum([ varpi[l,r][s] for r in range(0, k) ])
                                         for l in range(n)
                                         for s in (0, 1) if split[l] }

    # reconstructions
    for l in range(n):
      for r in range(k):
        v = sum([ coeffs[l,r,j] * f[-r+j] for j in range(nc) ])
        kernel.assign(fr[l,r], v)

    # weighted reconstruction
    for l in range(n):
      if not split[l]:
        v = sum([ omega[l,r] * fr[l,r] for r in range(k) ])
        if not self.weights_normalised:
           v /= sum([ omega[l,r] for r in range(k) ])
      else:
        vp = sum([ omega[l,r,0] * fr[l,r] for r in range(k) ])
        if not self.weights_normalised:
          vp /= sum([ omega[l,r,0] for r in range(k) ])
        vm = sum([ omega[l,r,1] * fr[l,r] for r in range(k) ])
        if not self.weights_normalised:
          vm /= sum([ omega[l,r,1] for r in range(k) ])
        v = scale[l,0] * vp - scale[l,1] * vm

      kernel.assign(fs[l], v)

    return kernel.body()


###############################################################################
# helpers

class Kernel(object):
  def __init__(self):
    import symbols
    if symbols.names.lang == 'fortran':
      self.code = codeprinters.FCodePrinter(settings={'source_format': 'free'})
    else:
      self.code = codeprinters.CCodePrinter()
    self.src = []
  def assign(self, dest, value):
    if isinstance(self.code, codeprinters.CCodePrinter):
      self.src.append(str(dest) + ' = ' + self.code.doprint(value.evalf(35)) + ';')
    else:
      self.src.append(str(dest) + ' = ' + self.code.doprint(value.evalf(35)))
  def body(self):
    return '\n'.join(self.src)



"""PyWENO symbol generators."""

import sympy

real = lambda x: sympy.Symbol(str(x), real=True)

class names(object):
  lang   = 'c'
  sigma  = 'sigma{}'
  omega  = 'omega{}r{}'
  f_star = 'fs{}'
  f_r    = 'fr{}r{}'
  f_mn   = 'f{:+d}{:+d}'
  f = {
    'c':       'f[(i{:+d})*fsi]',
    'opencl':  'f[(i{:+d})*fsi]',
    'fortran': 'f(i{:+d})'
    }

class SigmaGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'sigma')
    return real(tmp.format(idx))
  def all(self, k):
    return [ { 'r': r, 'name': self[r] } for r in range(0, k) ]

sigma = SigmaGenerator()

class OmegaGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'omega')
    if len(idx) == 3:
      return real(tmp.format(*idx[:2]) + ['p','m'][idx[2]])
    return real(tmp.format(*idx))
  def all(self, n, k, split):
    all = []
    for l in range(n):
      if not split[l]:
        for r in range(0, k):
          all.append({ 'l': l, 'r': r, 's': 0, 'pm': '', 'name': self[l,r] })
      else:
        for r in range(0, k):
          all.append({ 'l': l, 'r': r, 's': 0, 'pm': 'p', 'name': self[l,r,0] })
          all.append({ 'l': l, 'r': r, 's': 1, 'pm': 'm', 'name': self[l,r,1] })
    return all

omega = OmegaGenerator()

class FStarGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'f_star')
    return real(tmp.format(idx))
  def all(self, n):
    return [ { 'l': l, 'name': self[l] } for l in range(n) ]

fs = FStarGenerator()

class FRGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'f_r')
    return real(tmp.format(*idx))
  def all(self, n, k):
    return [ { 'l': l, 'r': r, 'name': self[l,r] } for l in range(n) for r in range(k) ]

fr = FRGenerator()

class FGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'f')[names.lang]
    return real(tmp.format(idx))

f = FGenerator()

class FMNGenerator(object):
  def __getitem__(self, idx):
    tmp = getattr(names, 'f_mn')
    return real(tmp.format(*idx).replace('-','m').replace('+','p'))

fmn = FMNGenerator()



"""PyWENO WENO reconstructor."""

import numpy as np
import pyweno.cweno

def reconstruct(q, k, points,
                n=None,
                smoothness=None,
                weights=None,
                return_smoothness=False,
                return_weights=False,
                squeeze=True):
  """Perform WENO reconstruction of q.

  :param q:                 cell-averaged unknown to reconstruct
  :param k:                 order of reconstruction (odd values from 5 to 11)
  :param points:            reconstruction points
  :param n:                 number of reconstruction points (optional)
  :param smoothness:        use smoothness indicators *smoothness* (optional)
  :param weights:           use non-linear weights *weights* (optional)
  :param return_smoothness: return smoothness indicators? (default: ``False``)
  :param return_weights:    return weights? (default: ``False``)
  :param squeeze:           squeeze the results? (default: ``True``)

  Supported reconstruction points *points* are:

  * ``'left'`` - left edge of each cell

  * ``'right'`` - right edge of each cell

  * ``'middle'`` - the middle of each cell

  * ``'+/-'`` - returns two reconstructions per cell, both at the left
    edge of each cell; the first is considered as a limit from the
    left (-), the second is considered as a limit from the right (+).

  * ``'gauss'`` or ``'gauss_legendre'`` - *n* point Gauss-Legendre
    quadrature points

  * ``'gauss_lobatto'`` - *n* point Gauss-Lobatto quadrature points

  * ``'gauss_radau'`` - *n* point Gauss-Radau quadrature points


  This function wraps several complied WENO kernels which were
  generated by PyWENO.  If you need WENO reconstructions at points not
  supported by this wrapper, please checkout `PyWENO
  <http://pyweno.readthedocs.org/>`_.

  """

  valid_k = (5, 11)
  valid_points = [ 'left', 'right', 'middle', '+/-',
                   'gauss', 'gauss_legendre',
                   'gauss_lobatto',
                   'gauss_radau' ]

  if (k % 2) == 0:
    raise ValueError('even order WENO reconstructions are not supported')

  if k < valid_k[0] or k > valid_k[1]:
    raise ValueError('%d order WENO reconstructions are not supported' % k)

  if not (points in valid_points):
    raise ValueError("'points' must be one of: " + ', '.join(valid_points))

  if points == 'gauss':
    points = 'gauss_legendre'

  N = q.shape[0]
  k = (k+1)/2

  # validate points and n
  if points in [ 'left', 'right', 'middle' ]:
    n = 1
  elif n is None:
    n = k

  if points in [ '+/-' ]:
    n = 2

  assert(n > 0)

  # smoothness
  if smoothness is None:
    smoothness = np.zeros((N,k))
    try:
      func = getattr(pyweno.cweno, 'smoothness%03d' % k)
    except AttributeError:
      raise ValueError('unsupported smoothness indicator: '
                       + 'k=%d not generated' % (2*k-1))
    func(q, smoothness)

  assert(smoothness.shape == (N,k))

  # weights
  if weights is None:
    weights = np.zeros((N,n,k))
    try:
      func = getattr(pyweno.cweno,
                     'weights_' + points + '%03d%03d' % (k, n))
    except AttributeError:
      raise ValueError('unsupported non-linear weights: '
                       + 'k=%d not generated' % (2*k-1))
    func(smoothness, weights)

  assert(weights.shape == (N,n,k))

  # reconstruct
  qr = np.zeros((N,n))

  if points == '+/-':
    # XXX: finish this off
    raise NotImplementedError, '+/- not implemented yet'

  try:
    func = getattr(pyweno.cweno,
                   'reconstruct_' + points + '%03d%03d' % (k, n))
  except AttributeError:
    raise ValueError('unsupported WENO reconstruction: '
                     + 'points=%s, k=%d, n=%d not generated'
                        % (points, 2*k-1, n))

  func(q, weights, qr)

  # done!
  if squeeze:
    qr = qr.squeeze()
    if weights is not None and return_weights:
      weights = weights.squeeze()

  if return_smoothness and return_weights:
    return (qr, smoothness, weights)

  if return_smoothness:
    return (qr, smoothness)

  if return_weights:
    return (qr, weights)

  return qr



"""PyWENO symbolic tool kit."""

import sympy
from itertools import product

from sympy import Symbol as sym


###############################################################################

def polynomial_interpolator(x, y):
  """Build a symbolic polynomial that interpolates the points (x_i, y_i).

  The returned polynomial is a function of the SymPy variable x.
  """

  k, xi = len(x), sym('x')

  sum_i = 0
  for i in range(k):
    ns = range(k)
    ns.remove(i)

    num, den = 1, 1
    for n in ns:
      num *= xi   - x[n]
      den *= x[i] - x[n]

    sum_i += num / den * y[i]

  return sum_i


###############################################################################

def primitive_polynomial_interpolator(x, y):
  """Build a symbolic polynomial that approximates the primitive
  function f such that f(x_i) = sum_j y_j * (x_{j+1} - x_{j}).

  Note: The x argument should be list that is one element longer
  than the y list.

  The returned polynomial is a function of the SymPy variable 'x'.
  """

  Y = [0]
  for i in range(len(y)):
    Y.append(Y[-1] + (x[i+1] - x[i]) * y[i])

  return polynomial_interpolator(x, Y)

ppi = primitive_polynomial_interpolator


###############################################################################

def _pt(a, b, x):
  """Map x in [-1, 1] to x in [a, b]."""

  half = sympy.sympify('1/2')

  w = half * (b - a)
  c = half * (a + b)

  return w * x + c


def reconstruction_coefficients(k, xi, d=0):
  r"""Compute the reconstruction coefficients for a 2k-1 order WENO
  scheme corresponding to the reconstruction points in *xi*.

  The reconstruction points in *xi* should in :math:`[-1, 1]`.  This
  interval is then mapped to the cell :math:`[x_{i-1/2}, x_{i+1/2}]`.

  The returned coefficients are stored as SymPy variables in a
  dictionary indexed according to ``c[l, r, j]``.  That is

  .. math::

    f^r(\xi_l) \approx \sum_j c^{l,r}_j \, f_{i-r+j}

  for each :math:`l` from 0 to ``len(xi)``.
  """

  i, n, x, dx  = k-1, len(xi), sym('x'), sym('dx')

  # build arrays of cell boundaries and cell averages
  xs = [ j*dx               for j in range(-k+1, k+1) ]
  fs = [ sym('f[i%+d]' % j) for j in range(-k+1, k)   ]

  c = { 'n': n, 'k': k, 'd': d }

  # compute reconstruction coefficients for each left shift r
  for l, r in product(range(n), range(k)):
    p = ppi(xs[i-r:i-r+k+1], fs[i-r:i-r+k]).diff(x, d+1)

    for j in range(k):
      z = _pt(xs[i], xs[i+1], xi[l])
      c[l, r, j] = p.subs(x, z).expand().coeff(fs[i-r+j])

      if c[l, r, j] is None:
        c[l, r, j] = 0

      c[l, r, j] *= dx**d

  return c


###############################################################################

def optimal_weights(k, xi, **kwargs):
  r"""Compute the optimal weights for a 2k-1 order WENO scheme
  corresponding to the reconstruction points in *xi*.

  The coefficients are stored as SymPy variables in a dictionary
  indexed according to ``w[l,r]``.  That is

  .. math::

    f(\xi^l) \approx \sum_r w^r f^r

  for each :math:`l` from 0 to ``len(xi)``.
  """

  n = len(xi)

  c   = reconstruction_coefficients(k, xi, **kwargs)
  c2k = reconstruction_coefficients(2*k-1, xi, **kwargs)

  omega = [ sym('omega%d' % r) for r in range(k) ]
  varpi = { 'n': n, 'k': k }
  split = { 'n': n }

  for l in range(n):

    # use first k eqns since weight system is overdetermined
    eqns = []
    for j in range(k):
      rmin, rmax = max(0, (k-1)-j), min(k-1, 2*(k-1)-j)
      terms = [ omega[r] * c[l, r, r-(k-1)+j] for r in range(rmin, rmax+1) ]
      eqns.append(sum(terms) - c2k[l, k-1, j])

    # XXX: Should make sure that when mpmath.mpf's are passed in
    # (in xi), that the solution obtained above is a high
    # precision solution.

    sol = sympy.solve(eqns, omega)

    # now check all 2*k-1 eqns to make sure the weights work out properly
    for j in range(2*k-1):
      rmin, rmax = max(0, (k-1)-j), min(k-1, 2*(k-1)-j)
      terms = [ omega[r] * c[l, r, r-(k-1)+j] for r in range(rmin, rmax+1) ]
      eqn = sum(terms) - c2k[l, k-1, j]
      err = eqn.subs(sol)
      if abs(err) > 1e-15:
        raise ValueError("optimal weight %d failed with error %s" % (j, err))

    # set weight or split as appropriate
    split[l] = min(sol.values()) < 0
    for r in range(k):
      if split[l]:
        w  = sol[omega[r]]
        wp = (w + 3*abs(w))/2
        wm = wp - w
        varpi[l, r] = (wp, wm)
      else:
        varpi[l, r] = sol[omega[r]]

  return varpi, split


###############################################################################

def jiang_shu_smoothness_coefficients(k):
  r"""Compute the Jiang-Shu smoothness coefficients for a 2k-1 order
  WENO scheme.

  The coefficients are stored as SymPy variables in a dictionary
  indexed according to ``beta[r, m, n]``.  That is

  .. math::

    \sigma^r = \sum_{m=1}^{2k-1} \sum_{n=1}^{2k-1}
      \beta_{r,m,n}\, \overline{f}_{i-k+m}\, \overline{f}_{i-k+n}.
  """

  x, dx, xi = sym('x'), sym('dx'), sym('x')

  # build arrays of cell boundaries and cell averages
  xs = [ j*dx               for j in range(-k+1, k+1) ]
  fs = [ sym('f[i%+d]' % j) for j in range(-k+1, k)   ]

  beta = { 'k': k }

  # compute smoothness coefficients for each left shift r
  for r in range(k):
    p = ppi(xs[k-1-r:2*k-r], fs[k-1-r:2*k-1-r]).diff(x)

    # sum of L^2 norms of derivatives
    s = 0
    for j in range(1, k):
      pp = sympy.diff(p, xi, j)**2
      pp = pp.as_poly(x)
      pp = pp.integrate(x)
      pp = dx**(2*j-1) * ( pp.subs(x, xs[k]) - pp.subs(x, xs[k-1]) )
      s += pp.expand()

    # pick out coefficients
    for m in range(k):
      for n in range(m, k):
        c = s.coeff(fs[k-1-r+m]*fs[k-1-r+n])
        if c is not None:
          beta[r, m, n] = c

  return beta


###############################################################################

def reconstruction_coefficients_for_derivative(k, bias):
  """XXX"""

  i = k-1
  (x, dx) = sympy.var('x dx')

  # build array of grid points (sympy vars x[i])
  xs = []
  for j in range(-k+1, k+1):
    xs.append(j*dx)

  # build array of point values (sympy vars f[i])
  fs = []
  for j in range(-k+1, k+1):
    fs.append(sympy.var('f[i%+d]' % j))

  n = 1

  # compute reconstruction coefficients for each left shift r
  c = { 'n': n, 'k': k, 'l': k+1 }

  for l in range(n):
    for r in range(0, k):
      p = polynomial_interpolator(xs[i-r:i-r+k+1], fs[i-r:i-r+k+1]).diff(x)

      z = xs[i] if bias=='+' else xs[i+1]

      for j in range(0, k+1):
        p = p.expand()
        c[l,r,j] = p.subs(x, z).coeff(fs[i-r+j])

        if c[l,r,j] is None:
          c[l,r,j] = 0 # i think it may be zero for k=2
        c[l,r,j] *= dx

  return c


###############################################################################

def optimal_weights_for_derivative(k, bias):
  """XXX"""

  n = 1
  c   = reconstruction_coefficients_for_derivative(k, bias)
  c2k = reconstruction_coefficients_for_derivative(2*k-1, bias)

  omega = []
  for r in range(k):
    omega.append(sympy.var('omega%d' % r))

  varpi = { 'n': n, 'k': k }
  split = { 'n': n }

  for l in range(n):
    eqns = []

    # we'll just use the first k eqns since the system is overdetermined
    for j in range(k):

      rmin = max(0, (k-1)-j)
      rmax = min(k-1, 2*(k-1)-j)

      accum = 0
      for r in range(rmin, rmax+1):
        accum = accum + omega[r] * c[l,r,r-(k-1)+j]

      eqns.append(accum - c2k[l,k-1,j])

    # XXX: Should make sure that when mpmath.mpf's are passed in
    # (in xi), that the solution obtained above is a high
    # precision solution.

    sol = sympy.solve(eqns, omega)

    # now check all 2*k-1 eqns to make sure the weights work out properly
    for j in range(2*k-1):

      rmin = max(0, (k-1)-j)
      rmax = min(k-1, 2*(k-1)-j)

      accum = 0
      for r in range(rmin, rmax+1):
        accum = accum + omega[r] * c[l,r,r-(k-1)+j]

      eqn = accum - c2k[l,k-1,j]
      eqn.subs(sol)             # XXX: check this

    # check for negative weights and mark as split
    if min(sol.values()) < 0:
      split[l] = True
    else:
      split[l] = False

    # split as appropriate
    for r in range(k):
      if split[l]:
        w  = sol[omega[r]]
        wp = (w + 3*abs(w))/2
        wm = wp - w
        varpi[l,r] = (wp, wm)
      else:
        varpi[l,r] = sol[omega[r]]

  return (varpi, split)


###############################################################################

def jiang_shu_smoothness_coefficients_for_derivative(k, bias):
  """XXX"""

  (x, dx) = sympy.var('x dx')

  # build array of grid points (sympy vars x[i])
  xs = []
  for j in range(-k+1, k+1):
    xs.append(j*dx)

  # build array of point values (sympy vars f[i])
  fs = []
  for j in range(-k+1, k+1):
    fs.append(sympy.var('f[i%+d]' % j))

  # compute reconstruction coefficients for each left shift r
  beta = { 'k': k, 'l' : k+1 }
  for r in range(0, k):
    p = polynomial_interpolator(xs[k-1-r:2*k-r], fs[k-1-r:2*k-r])
    # sum of L^2 norms of derivatives
    s = 0
    for j in range(2, k+1):
      # start at the second derivative since we are
      # interested in the smoothness of the derivative of p
      pp = (sympy.diff(p, x, j))**2
      pp = sympy.integrate(pp, x)
      q = (-1 if bias=='+' else 1) * dx/2
      pp = (xs[k] - xs[k-1])**(2*j-1) * (
        pp.subs(x, xs[k]) - pp.subs(x, xs[k-1]) )
      pp = pp.expand()
      s = s + pp

    # pick out coefficients
    for m in range(k+1):
      for n in range(m, k+1):
        c = s.coeff(fs[k-1-r+m]*fs[k-1-r+n])
        if c is not None:
          beta[r,m,n] = c

  return beta




"""PyWENO version information.

   To obtain the version of PyWENO::

     >>> import pyweno.version
     >>> pyweno.version.version()
     >>> pyweno.version.git_version()

   """

import imp
import os

def _version(name):
    version_file = os.path.join(os.path.dirname(__file__), name + '.py')
    version = imp.load_module('pyweno.pyweno.' + name,
                              open(version_file),
                              version_file,
                              ('.py', 'U', 1))
    return version.version

def version():
    """Return current version."""
    return _version('__version__')

def git_version():
    """Return current *git* version (if available)."""
    return _version('__git_version__')



"""PyWENO non-uniform reconstruction coefficients."""

import numpy as np
import pyweno.ccoeffs


def reconstruction_coeffs(xi, i, r, k, x):
  """Numerically compute the reconstruction coefficients.

  The reconstruction coefficients *c_j* are used to approximate the
  function *f* given its cell averages *f_j* according to

  .. math::

    f(\\xi) \\approx \sum_{j=0}^{k-1} c_j\, \overline{f}_{i-r+j}.

  :param xi: reconstruction point
  :param i: cell index (ie, index of the cell that containts *xi*)
  :param r: left shift
  :param k: order
  :param x: cell boundaries


  """

  c = np.empty(k)

  try:
    pyweno.ccoeffs.reconstruction_coeffs(xi, i, r, k, x, c)
  except:
    raise NotImplementedError, "reconstruction coeffs for k = %d not implemented yet" % (k)

  return c



"""PyWENO non-uniform reconstruction routines."""

import numpy as np

def coefficients(k, xi, x):
    """Numerically compute reconstruction and smoothness coefficients.

    This wraps the "symbolic" C routines in the cnonuniform extension
    module.

    XXX: Need to implement: higher derivatives, optimal weights.

    XXX: When the above are done, can probably replace all nonuniform
    stuff below with this routine.
    """

    import pyweno.cnonuniform

    x  = np.asarray(x,  np.float64)
    xi = np.asarray(xi, np.float64)

    nc = len(x) - 1
    n  = len(xi)
    c     = np.zeros((nc, n, k, k), np.float64)
    beta  = np.zeros((nc, k, k, k), np.float64)
    varpi = np.zeros((nc, n, k), np.float64)

    pyweno.cnonuniform.nonuniform_coeffs(k, xi, x, c, beta, varpi)

    return c, beta, varpi



'''PyWENO setup script.'''

import glob
import os
import re

from distutils.core import setup
from distutils.extension import Extension

import numpy as np


###############################################################################
# version

execfile('version.py')                  # this sets 'version'

with open('README', 'r') as file:
    long_description = file.read()


###############################################################################
# save git version to 'pyweno/__git_version__.py'

try:
    git_head_file = os.path.join(os.path.dirname(__file__), '.git', 'HEAD')
    f = open(git_head_file)
    m = re.match(r'ref: (.+)', f.readline())
    ref = m.group(1)
    f.close()

    git_head_file = os.path.join(os.path.dirname(__file__), '.git', ref)
    f = open(git_head_file)
    git_version = f.readline().rstrip()
    f.close()

except:
    git_version = 'not_available'

git_version_file = os.path.join(os.path.dirname(__file__),
                                'pyweno','__git_version__.py')
f = open(git_version_file, 'w')
f.write("version = '%s'\n" % (git_version))
f.close()


###############################################################################
# save version to 'pyweno/__version__.py'

version_file = os.path.join(os.path.dirname(__file__),
                            'pyweno','__version__.py')
f = open(version_file, 'w')
f.write("version = '%s'\n" % (version))
f.close()


###############################################################################
# setup!

setup(

    name         = "PyWENO",
    packages     = ['pyweno'],
    version      = version,
    author       = "Matthew Emmett",
    author_email = "matthew@emmett.ca",
    description  = "Weighted Essentially Non-oscillatory (WENO) reconstructions.",
    license      = "BSD",
    keywords     = "weno, interpolate, interpolation, finite, volume",
    url          = "http://readthedocs.org/docs/pyweno/en/latest/",

    ext_modules = [
        Extension('pyweno.ccoeffs',
                  sources = ['src/ccoeffs.c'] + glob.glob('src/coeffs*.c'),
                  include_dirs=[np.get_include()]
                  ),
        Extension('pyweno.cweno',
                  sources = ['src/cweno.c'] + glob.glob('src/weno*.c'),
                  include_dirs=[np.get_include()],
                  extra_compile_args = ['-std=c99'],
                  ),
        Extension('pyweno.cnonuniform',
                  sources = ['src/nfweno.c', 'src/poly.c'],
                  include_dirs=[np.get_include()],
                  extra_compile_args = ['-std=c99'],
                  )],

    long_description = long_description,

    classifiers = [
        'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
#        'Development Status :: 6 - Mature',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python ',
        'Programming Language :: C',
        'Programming Language :: Fortran',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        ],

    )



"""PyWENO code generation helpers."""

from sympy.printing.fcode import FCodePrinter as SympyFCodePrinter
from sympy.printing.ccode import CCodePrinter as SympyCCodePrinter
from sympy.printing.precedence import precedence


class FCodePrinter(SympyFCodePrinter):
  pass

class CCodePrinter(SympyCCodePrinter):
  def _print_Pow(self, expr):
    if expr.exp == 2:
      PREC = precedence(expr)
      s = str(self.parenthesize(expr.base, PREC))
      return '%s*%s' % (s,s)
    else:
      return super(CCodePrinter,self)._print_Pow(expr)
